#!/usr/bin/env python
"""
simulates a system of authors and research papers
see README and http://arxiv.org/abs/??? for details
matus.medo@unifr.ch, June 2016
"""


import numpy as np
from numpy.random import seed, rand, exponential, randint, choice
from scipy.stats import pearsonr, sem
from os.path import isdir
from os import makedirs
import sys


REALIZATIONS = 100              # number of model realizations to produce
R = 1000                        # number of researchers in the system
evolve_researcher_group = True # whether some researchers become active later (if True) or everyone is active from the beginning (if False)
frac_later_start = 0.75         # the fraction of researchers who start being active later than in step 0
TMAX = 10                       # simulation length in years
lifetime = 2                    # lifetime of each paper in years
A = 1                           # initial attractiveness of each paper
kO = 20                         # outdegree of each paper
one_author_per_paper = False    # whether to assume that each paper has exactly one author or whether to use a real distribution of the number of co-authors
how_get_fitness = 'max'         # fitness of a paper is obtained as maximum over ability of all co-authors (max), average over all co-authors (avg), or randomly chosen from all co-authors (rand)
fitness_noise = 0.2             # the level of noise which applies to the author ability in order to obtain paper quality
ab_add = 1.0                    # additional term for the researcher ability distribution
fRAND = 0.0                     # fraction of citations that are pointed at random
max_coauthors = 100             # the maximal number of co-authors of a paper
P_max = R * 100                 # maximal allowed number of papers (we don't expect to simulate systems where an average researcher has more than 100 papers)
P_max_ind = 50 * TMAX           # maximal allowed number of papers for an individual researcher (50 papers per year)
skip_Omega_control = True       # whether we do skip the control of initial low Omega values or whether we know the stationary Omega for given parameters and thus we do not have to skip the control
degree_fit_threshold = 50       # paper degree distributions are fit with a power-law with this lower bound
seed_val = 0                    # seed for the RNG (unimportant, set only for the sake of reproducibility)
output_dir = 'sim_data'         # directory, in which simulation data (networks, top-ranked researchers, etc.) are saved

R_abil = np.zeros(R)                                            # researcher ability
R_act = np.zeros(R, dtype = int)                                # researcher activity
R_potential = np.zeros(R, dtype = int)                          # researcher potential
R_deg = np.zeros(R, dtype = int)                                # researcher degree
R_pub = np.zeros((R, P_max_ind), dtype = int)                   # list of papers published by each researcher
R_start = np.zeros(R, dtype = int)                              # the month in which the author becomes active (i.e., in which he is first included in the choice of authors)
R_hirsch = np.zeros(R, dtype = int)                             # researcher h-index values (stored because they are costly to compute and used by several metrics)
P_deg = np.zeros(P_max, dtype = int)                            # paper degree
P_tau = np.zeros(P_max, dtype = int)                            # paper birth time (in months)
P_authnum = np.zeros(P_max, dtype = int)                        # paper number of authors
P_relI = np.zeros(P_max)                                        # paper initial relevance
P_decay = np.ones(P_max)                                        # paper time decay factor
P_authors = np.zeros((P_max, max_coauthors), dtype = int)       # paper authors
P_references = np.zeros((P_max, kO), dtype = int)               # referred papers
Omega = np.zeros(12 * TMAX + 1)                                 # normalization term Omega(t)
Omega_final = []                                                # a list with final Omega values
powerlaw_exp = []                                               # a list with fitted power-law exponent values of the paper degree distribution
P_pass_threshold = []                                           # a list with the number of papers that pass the degree threshold for fitting the power-law distribution
final_P = []                                                    # a list with final number of papers


def first_init():                                               # the first initialization when the file with stationary Omega values is loaded and parsed
    global Omega_stationary, skip_Omega_control

    seed(seed_val)                                              # initialize the RNG
    questionmark_line_found = False                             # search for a Omega stationary value corresponding to the current parameter setting
    f = open('Omega_stationary.dat', 'r')
    for line in f:
        if line[0] <> '#':
            parts = line.split('\t')
            if (int(parts[0].split('=')[1]) == R and int(parts[1].split('=')[1]) == TMAX and float(parts[2].split('=')[1]) == lifetime and
                float(parts[3].split('=')[1]) == A and float(parts[4].split('=')[1]) == kO and str2bool(parts[5].split('=')[1]) == one_author_per_paper and
                parts[6].split('=')[1] == how_get_fitness and float(parts[7].split('=')[1]) == fitness_noise and
                float(parts[8].split('=')[1]) == ab_add and float(parts[9].split('=')[1]) == fRAND and
                str2bool(parts[10].split('=')[1]) == evolve_researcher_group):
                try: Omega_stationary = float(parts[11].split('=')[1])
                except:
                    questionmark_line_found = True
                    break
                print 'found a matching parameter setting and loaded the stationary Omega value {:.2e}'.format(Omega_stationary)
                if skip_Omega_control == False: print 'oops, it seems that this is not the only entry for the given parameter setting'
                skip_Omega_control = False
    f.close()
    if skip_Omega_control:
        print '    !!!! do not take the simulation results too seriously (especially the popularity of old papers) because the Omega_stationary was not found'
        if not questionmark_line_found:
            f = open('Omega_stationary.dat', 'a')
            f.write('R={}\tTMAX={}\tlifetime={}\tA={}\tkO={}\tone_auth_per_pap={}\thow_get_fitness={}\tfitness_noise={}\tab_add={}\tfRAND={}\tevolve_auth={}\tstatOmega=???\n'.format(R, TMAX, lifetime, A, kO, one_author_per_paper, how_get_fitness, fitness_noise, ab_add, fRAND, evolve_researcher_group))
            f.close()

    try: makedirs(output_dir)                                   # try to create a results directory
    except:
        if not isdir(output_dir):
            print 'oops, something went wrong - failed to create the results directory "{}"'.format(output_dir)
            sys.exit(1)


def str2bool(v):
  return str(v).lower() in ("yes", "true", "t", "1")


def coauthor_distrib(n):                # the fraction of all papers that are authored by n co-authors (obtained by fitting the MAS data)
    return n / (100 + np.power(n, 4.6))


def activity_distrib(n):                # the fraction of all researchers that authored n papaers (obtained by fitting the MAS data)
    return np.exp(-(np.log(n)+3.5)**2 / 5.9)


def init_relevance_paper(authors):      # initial relevance of a paper authored by authors
    if how_get_fitness == 'max':
        base_fitness = np.max(R_abil[authors])
    elif how_get_fitness == 'avg':
        base_fitness = np.mean(R_abil[authors])
    elif how_get_fitness == 'rand':
        base_fitness = choice(R_abil[authors])
    else:
        print 'how_get_fitness {} undefined'.format(how_get_fitness)
        exit(1)
    return base_fitness * (1 + fitness_noise * (2 * rand() - 1))        # return the base fitness modulated with noise


def get_researcher_ability():           # generate ability values for all researchers
    return ab_add + exponential(1, size = R)


def get_researcher_activity():          # generate activity values for all researchers
    act_range = np.arange(1, P_max_ind + 1)
    temp = choice(act_range, size = R, p = activity_distrib(act_range) / activity_distrib(act_range).sum())
    return (temp * float(TMAX) / 10).astype(int)                        # if simulation runs for a different period of time than 10, return correspondingly higher/smaller numbers


def get_researcher_start():             # generate the activity start for all researchers
    values = np.zeros(R)
    for r in range(R):
        if rand() < frac_later_start:   # some researchers start later
            values[r] = randint(1, high = 12 * TMAX - 36)               # note that all researchers are in the system for at least 3 years (36 months)
    return values


def modify_act():                       # proportionally decrease the activity of researchers who start late
    global R_act

    for r in range(R):
        R_act[r] = int(R_act[r] * (1 - float(R_start[r]) / (12 * TMAX)))
        if R_act[r] == 0: R_act[r] = 1


def get_number_of_authors():            # generate the number of authors for a paper
    if one_author_per_paper: return 1
    coauth_range = np.arange(1, max_coauthors + 1)
    coauthor_norm = coauthor_distrib(coauth_range).sum()
    return choice(coauth_range, p = coauthor_distrib(coauth_range) / coauthor_distrib(coauth_range).sum())


def init(realization):                                                  # initialize the simulation arrays
    global P, R_abil, R_act, R_potential, R_deg, R_rep, R_pub, P_deg, P_tau, P_relI, P_decay, P_authors, Omega

    P = 0                                                       # no papers at the beginning
    R_abil[:] = get_researcher_ability()                        # researcher ability
    R_act[:] = get_researcher_activity()                        # researcher activity
    if evolve_researcher_group:
        R_start[:] = get_researcher_start()                     # get the time step in which researcher activity starts
        modify_act()                                            # modify the activity values correspondingly (late comers have their activity proportionally reduced)
    else:
        R_start[:] = 0                                          # if the group of researchers is fixed, everyone starts in time step 0
    R_potential[:] = 0                                          # researcher potential is originally zero and then set to its starting value at time step R_start
    R_deg[:] = 0                                                # researcher degree
    R_pub[:] = 0                                                # list of papers published by each researcher
    P_deg[:] = 0                                                # paper degree
    P_tau[:] = 0                                                # paper birth time
    P_authnum[:] = 0                                            # paper number of authors
    P_relI[:] = 0                                               # paper initial relevance
    P_decay[:] = 1                                              # paper time decay factor
    P_authors[:, :] = -1                                        # paper authors
    P_references[:, :] = -1                                     # paper references
    Omega[:] = 0                                                # normalization term Omega(t)
    print 'in {} years, the least active researcher is expected to produce {:.1f} papers'.format(TMAX, np.min(R_act))
    print 'in {} years, the average researcher is expected to produce {:.1f} papers'.format(TMAX, np.mean(R_act))
    print 'in {} years, the most active researcher is expected to produce {:.1f} papers'.format(TMAX, np.max(R_act))
    if np.sum(R_act) > 0.9 * P_max:
        print 'the expected total number of papers, {:.0f}, is very close to the pre-set maximal number of papers, {}'.format(np.sum(R_potential), P_max)
        sys.exit(1)


def choose_target(weights):             # choose one item proportionally to the given item weigths
    normw = weights / weights.sum()
    x = rand()                          # choose a random number in [0, 1)
    n = len(weights)                    # go from the last item to the first
    while x > 0 and n > 0:              # this returns n with probability given by weights
        n -= 1
        x -= normw[n]
    return n


def add_paper(t, authors, Omega, weights):                              # authors add a new paper at time t
    global P, P_rel, P_relI, P_deg, P_author, R_deg

    P_tau[P] = t                                                        # set the paper's birth time
    P_authnum[P] = len(authors)                                         # set the paper's number of authors
    P_relI[P] = init_relevance_paper(authors)                           # generate the initial relevance of the new paper with these authors
    P_authors[P,:len(authors)] = authors                                # set authors to be the authors of paper P
    R_deg[authors] += 1
    links_to_make = kO
    if links_to_make > weights.size:                                    # if there are fewer target papers available than kO, lower links_to_make
        links_to_make = weights.size
    linked = []                                                         # papers linked so far
    while links_to_make > 0:                                            # create kO references (links from the new paper to the existing papers)
        if skip_Omega_control or rand() < Omega / Omega_stationary:     # compare with the assumed stationary value of Omega
            if rand() < fRAND:                                          # choose the target paper at random
                while True:
                    target = randint(len(weights))
                    if target not in linked: break                      # if the chosen target has not been linked by the new paper, add accept it
            else:                                                       # choose the target paper according to the weight vector
                while True:
                    target = choose_target(weights)
                    if target not in linked: break                      # if the chosen target has not been linked by the new paper, add accept it
            P_references[P, len(linked)] = target                       # store all the references of paper P
            linked.append(target)                                       # add the target to the list of papers already linked by the new paper
            P_deg[target] += 1                                          # augment the target paper's degree
        links_to_make -= 1
    P += 1                                                              # augment the number of papers


def save_output(realization):
    global powerlaw_exp, P_pass_threshold

    log_P_deg = np.zeros(P)
    for p in range(P):
        if P_deg[p] > 0: log_P_deg[p] = np.log(P_deg[p])
    print 'Pearson correlations are r(P_deg, P_initial_relevance) = {:.4f} and r(log(P_deg), P_initial_relevance) = {:.4f}'.format(pearsonr(P_deg[:P], P_relI[:P])[0], pearsonr(log_P_deg, P_relI[:P])[0])

    above_threshold = np.where(P_deg >= degree_fit_threshold)[0]
    exp_val = 1 + above_threshold.size / np.sum(np.log(P_deg[above_threshold] / float(degree_fit_threshold)))
    print 'power-law exponent of the paper degree distribution: {:.2f} ({} papers pass the degree fit threshold {})'.format(exp_val, above_threshold.size, degree_fit_threshold)
    powerlaw_exp.append(exp_val)
    P_pass_threshold.append(above_threshold.size)
    print 'running average of the power-law exponents: {:.2f} ({:.0f} papers pass the threshold on average)'.format(np.mean(np.array(powerlaw_exp)), np.mean(np.array(P_pass_threshold)))

    g = open('{}/{}-authors.dat'.format(output_dir, realization), 'w')
    g.write('# author\tactivity\tability\n')
    for author in range(R):
        g.write('{}\t{}\t{:.6f}\n'.format(author, R_deg[author], R_abil[author]))
    g.close()

    g = open('{}/{}-papers.dat'.format(output_dir, realization), 'w')
    g.write('# paper\tpublished_at\tinitial_fitness\tauthors\n')
    for p in range(P):
        g.write('{}\t{}\t{:.6f}\t'.format(p, P_tau[p], P_relI[p]))
        for author in P_authors[p]:
            if author <> -1: g.write('{};'.format(author))
        g.write('\n')
    g.close()

    g = open('{}/{}-references.dat'.format(output_dir, realization), 'w')
    g.write('# citing_paper\tcited_paper\n')
    for p in range(P):
        for ref in P_references[p]:
            if ref <> -1: g.write('{}\t{}\n'.format(p, ref))
    g.close()


def simulate():
    global P, R_potential                                               # current number of papers

    num_papers_added = 0
    for o in range(REALIZATIONS):
        print '\nsimulation #{}'.format(o + 1)
        init(o)                                                         # initialize everything        
        for t in range(12 * TMAX):                                      # simulation runs in months
            for r in range(R):                                          # researcher potential is initialized at the time step when the researcher becomes active
                if R_start[r] == t: R_potential[r] = R_act[r]
            removed_this_month = float(np.sum(R_potential)) / (12 * TMAX - t)   # how much should be substracted from the R_potential array in this month
            if P > 0:
                pap_attr = (A + P_deg[:P]) * P_relI[:P] * P_decay[:P]   # compute the array of paper attractiveness
            else: pap_attr = np.array([])
            Omega[t] = np.sum(pap_attr)
            potential_removed = 0
            while potential_removed < removed_this_month:               # in each month, we remove in_one_month from the author potential array
                num_coauthors = get_number_of_authors()
                if np.where(R_potential > 0)[0].size >= num_coauthors:  # if there are enough options, choose the authors based on R_potential
                    authors = choice(R, size = num_coauthors, replace = False, p = R_potential.astype(float) / np.sum(R_potential))
                elif t < 12 * TMAX - 1:                                 # if there are not enough researchers available and there will be more simulation steps in the future, stop
                    print 'stopped when trying to add a paper with {} authors (number of researchers with positive potential is {})'.format(num_coauthors, np.where(R_potential > 0)[0].size)
                    break
                else:
                    authors = np.where(R_potential > 0)[0]              # if it's the last step, then simply take all the remaining users with positive potential
                add_paper(t, authors, Omega[t], pap_attr)               # add one paper by authors assuming the attractiveness vector pap_attr
                R_potential[authors] -= 1
                potential_removed += num_coauthors
                num_papers_added += 1
            P_decay[:P] *= np.exp(-float(1)/(12 * lifetime))            # if this is done each month, relevance of all papers decays as R_p * exp(-age / lifetime)
            if (t + 1) % 12 == 0:
                print 'sim {}, year {:02d}: {} new papers, {} papers in total, most popular has {} citations (paper {}, fitness {:.2f}); average citations {:.1f}; Omega {:.2e}'.format(o + 1, (t + 1) / 12, num_papers_added, P, np.max(P_deg[:P]), np.argmax(P_deg[:P]), P_relI[np.argmax(P_deg)], np.mean(P_deg[:P]), Omega[t])
                num_papers_added = 0
        Omega[12 * TMAX] = np.sum((A + P_deg[:P]) * P_relI[:P] * P_decay[:P])
        Omega_final.append(Omega[12 * TMAX])
        print 'final Omega value: {:.2e}'.format(Omega[12 * TMAX])
        if skip_Omega_control: print 'running average of the final Omega values: {:.2e} (assumed stationary value is None)'.format(np.mean(np.array(Omega_final)))
        else: print 'running average of the final Omega values: {:.2e} (assumed stationary value is {:.2e})'.format(np.mean(np.array(Omega_final)), Omega_stationary)
        final_P.append(P)
        save_output(o)


first_init()
simulate()

g = open('overall_stats.txt', 'a')
print '\n\nTMAX={}; lifetime={}; ab_add={}; base_fitness={}; fitness_noise={}; fRAND={}; evolve_researcher_group={}'.format(TMAX, lifetime, ab_add, how_get_fitness, fitness_noise, fRAND, evolve_researcher_group)
g.write('TMAX={}; lifetime={}; ab_add={}; base_fitness={}; fitness_noise={}; fRAND={}; evolve_researcher_group={}\n'.format(TMAX, lifetime, ab_add, how_get_fitness, fitness_noise, fRAND, evolve_researcher_group))
Om_f = np.array(Omega_final)
if skip_Omega_control: print '\n\nfinal omega values in {} realizations: {:.2e} +- {:.2e} (assumed stationary value was None)'.format(REALIZATIONS, np.mean(Om_f), sem(Om_f))
else:
    print 'final omega values in {} realizations: {:.2e} +- {:.2e} (assumed stationary value was {:.2e})'.format(REALIZATIONS, np.mean(Om_f), sem(Om_f), Omega_stationary)
    g.write('final omega values in {} realizations: {:.2e} +- {:.2e} (assumed stationary value was {:.2e})\n'.format(REALIZATIONS, np.mean(Om_f), sem(Om_f), Omega_stationary))
pl_exp = np.array(powerlaw_exp)
print 'fitted power-law exponent values in {} realizations: {:.2f} +- {:.2f} (fitting lower bound {})'.format(REALIZATIONS, np.mean(pl_exp), sem(pl_exp), degree_fit_threshold)
g.write('fitted power-law exponent values in {} realizations: {:.2f} +- {:.2f} (fitting lower bound {})\n\n'.format(REALIZATIONS, np.mean(pl_exp), sem(pl_exp), degree_fit_threshold))
f_P = np.array(final_P)
print 'average number of papers in one simulation: {:.0f} +- {:.0f}'.format(np.mean(f_P), sem(f_P))
g.close()
